"""
Various functions to deal with conversion mpz <-> Python int/long

For doctests, see :class:`Integer`.

AUTHORS:

- Gonzalo Tornaria (2006): initial version

- David Harvey (2007-08-18): added ``mpz_get_pyintlong`` function
  (:trac:`440`)

- Jeroen Demeyer (2015-02-24): moved from c_lib, rewritten using
  ``mpz_export`` and ``mpz_import`` (:trac:`17853`)
"""

#*****************************************************************************
#       Copyright (C) 2015 Jeroen Demeyer <jdemeyer@cage.ugent.be>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************


from cpython.int cimport PyInt_FromLong, PyInt_Check
from cpython.long cimport PyLong_CheckExact, PyLong_FromLong
from mpz cimport *

cdef extern from *:
    ctypedef long Py_hash_t

cdef extern from "longintrepr.h":
    cdef _PyLong_New(Py_ssize_t s)
    cdef long PyLong_SHIFT
    ctypedef unsigned int digit
    ctypedef struct PyLongObject:
        Py_ssize_t ob_size
        digit* ob_digit

# Unused bits in every PyLong digit
cdef size_t PyLong_nails = 8*sizeof(digit) - PyLong_SHIFT


cdef mpz_get_pylong_large(mpz_srcptr z):
    """
    Convert a non-zero ``mpz`` to a Python ``long``.
    """
    cdef size_t nbits = mpz_sizeinbase(z, 2)
    cdef size_t pylong_size = (nbits + PyLong_SHIFT - 1) // PyLong_SHIFT
    L = _PyLong_New(pylong_size)
    mpz_export((<PyLongObject*>L).ob_digit, NULL,
            -1, sizeof(digit), 0, PyLong_nails, z)
    if mpz_sgn(z) < 0:
        (<PyLongObject*>L).ob_size = -(<PyLongObject*>L).ob_size
    return L


cdef mpz_get_pylong(mpz_srcptr z):
    """
    Convert an ``mpz`` to a Python ``long``.
    """
    if mpz_fits_slong_p(z):
        return PyLong_FromLong(mpz_get_si(z))
    return mpz_get_pylong_large(z)


cdef mpz_get_pyintlong(mpz_srcptr z):
    """
    Convert an ``mpz`` to a Python ``int`` if possible, or a ``long``
    if the value is too large.
    """
    if mpz_fits_slong_p(z):
        return PyInt_FromLong(mpz_get_si(z))
    return mpz_get_pylong_large(z)

cdef int mpz_set_pyintlong(mpz_ptr z, val) except -1:
    if PyInt_Check(val):
        return mpz_set_pylong(z, long(val))
    elif PyLong_Check(val):
        return mpz_set_pylong(z, val)
    else:
        return -1

cdef int mpz_set_pylong(mpz_ptr z, L) except -1:
    """
    Convert a Python ``long`` `L` to an ``mpz``.
    """
    cdef Py_ssize_t pylong_size = (<PyLongObject*>L).ob_size
    if pylong_size < 0:
        pylong_size = -pylong_size
    mpz_import(z, pylong_size, -1, sizeof(digit), 0, PyLong_nails,
            (<PyLongObject*>L).ob_digit)
    if (<PyLongObject*>L).ob_size < 0:
        mpz_neg(z, z)


cdef Py_hash_t mpz_pythonhash(mpz_srcptr z):
    """
    Hash an ``mpz``, where the hash value is the same as the hash value
    of the corresponding Python ``long``.
    """
    # Add all limbs, adding 1 for every carry
    cdef mp_limb_t h1 = 0
    cdef mp_limb_t h0
    cdef size_t i, n
    n = mpz_size(z)
    for i in range(n):
        h0 = h1
        h1 += mpz_getlimbn(z, i)
        # Add 1 on overflow
        if h1 < h0: h1 += 1

    cdef Py_hash_t h = h1
    if mpz_sgn(z) < 0:
        h = -h
    if h == -1:
        return -2
    return h



# -*- coding: utf-8 -*-

from setuptools import setup
#from distutils.extension import Extension
from setuptools import Extension
from Cython.Distutils import build_ext

with open('README') as f:
    long_desc = f.read()

classif = ['Development Status :: 3 - Alpha',
           'Intended Audience :: Developers',
           'Intended Audience :: Science/Research',
           'Topic :: Scientific/Engineering :: Mathematics',
           'License :: OSI Approved :: GNU General Public License (GPL)',
           'Programming Language :: Python :: 2.7']

setup(
    name="pyparma",
    version="0.3.3",
    description="Bindings to the parma polyhedra library,\
                 allowing to use double description from Python",
    long_description=long_desc,
    url="https://github.com/haudren/pyparma",
    author="Hervé Audren",
    author_email="h.audren@aist.go.jp",
    license="GPL",
    classifiers=classif,
    packages=["pyparma", "pyparma/utils"],
    cmdclass={'build_ext': build_ext},
    ext_modules=[
        Extension("pyparma.pylong",
            sources=["pyparma/pylong.pyx"],
            libraries=["gmp", "gmpxx"]
                  ),
        Extension("pyparma.ppl",
            sources=["pyparma/ppl.pyx",
                     "pyparma/ppl_shim.cc"],
            libraries=["ppl", "gmpxx"],
            language="c++",
            extra_compile_args=["-fPIC"]
                  )
                ],
    package_data={'pyparma': ['*.pxd', '*.hh', '*.cc', '*.cpp']}
)



"""
Various functions to deal with conversion mpz <-> Python int/long
"""
from types cimport *

cdef extern from *:
    ctypedef long Py_hash_t

cdef mpz_get_pylong(mpz_srcptr z)
cdef mpz_get_pyintlong(mpz_srcptr z)
cdef int mpz_set_pylong(mpz_ptr z, L) except -1
cdef int mpz_set_pyintlong(mpz_ptr z, val) except -1
cdef Py_hash_t mpz_pythonhash(mpz_srcptr z)
